////////////////////////////////////////////////////////////////////////////////////
//
// This Code Is Licensed Under The Xopus Open Source License
// More information: http://xopus.org, info@xopus.org
// Mailing list: http://groups.yahoo.com/group/xopus
//
// This product includes software developed by Q42 BV (http://www.q42.nl/).
// please visit http://www.q42.nl for more information or mail us: xopus@q42.nl
//
////////////////////////////////////////////////////////////////////////////////////

function ieDOMRange()
{
  this.startContainer = null;
  this.startOffset    = 0;
  this.endContainer   = null;
  this.endOffset      = 0;

  this.collapsed = true;
  this.commonAncestorContainer = null;
  this.ownerDocument = null;
}

ieDOMRange.prototype =
{
  /**
   * Set the start point of the range.
   */
  setStart: function(parent, offset)
  {
    this.startContainer = parent;
    this.startOffset    = offset;

    this.recalcCommonAncestor();
  },

  /**
   * Set the end point of the range.
   */
  setEnd: function(parent, offset)
  {
    this.endContainer = parent;
    this.endOffset    = offset;

    this.recalcCommonAncestor();
  },

  /**
   * Calculate the common ancestor of both the start and end point.
   */
  recalcCommonAncestor: function()
  {
    if (this.startContainer == null || this.endContainer == null)
      return;

    for (var startAncestor = this.startContainer; startAncestor; startAncestor = startAncestor.parentNode)
    {
      for (var endAncestor = this.endContainer; endAncestor; endAncestor = endAncestor.parentNode)
      {
        if (startAncestor == endAncestor)
        {
          this.commonAncestorContainer = startAncestor;
          this.ownerDocument = this.commonAncestorContainer.ownerDocument || this.commonAncestorContainer.parentNode.ownerDocument;
          return;
        }
      }
    }
  },

  /**
   * Remove the contents of the current selection and return them in a DocumentFragment
   */
  extractContents: function()
  {
    var df = this.cloneContents();
    this.deleteContents();
    return df;
  },

  /**
   * Delete the contents of the current selection.
   */
  deleteContents: function()
  {
    var leftSplit  = this.splitTree(this.commonAncestorContainer, this.startContainer, this.startOffset);
    var rightSplit = this.splitTree(this.commonAncestorContainer, this.endContainer, this.endOffset);

    xopus_globs.utils.removeChildren(this.commonAncestorContainer);

    var newContents = this.excludeSets(leftSplit, rightSplit);
    if (this.commonAncestorContainer.nodeType == 3)
    {
      this.commonAncestorContainer.nodeValue = newContents.nodeValue;

      // Collapse range
      this.startContainer = this.commonAncestorContainer;
      this.endContainer = this.commonAncestorContainer;

      this.startOffset = leftSplit.idx;
      this.endOffset = leftSplit.idx;
    }
    else
    {
      this.commonAncestorContainer.appendChild(newContents);

      // Collapse range
      this.startContainer = this.commonAncestorContainer;
      this.endContainer = this.commonAncestorContainer;

      this.startOffset = leftSplit.left.childNodes.length;
      this.endOffset = leftSplit.left.childNodes.length;
    }
  },

  /**
   * Get a DocumentFragment containing the contents of the current selection.
   */
  cloneContents: function()
  {
    var leftSplit  = this.splitTree(this.commonAncestorContainer, this.startContainer, this.startOffset);
    var rightSplit = this.splitTree(this.commonAncestorContainer, this.endContainer, this.endOffset);

    return this.intersectSets(leftSplit, rightSplit);
  },

  /**
   * Split the given subtree, in a left and right side.
   * root        is the root of the tree
   * leaf        is a node in which to split,
   * leafOffset  is either the character offset in a text node or
   *             the index of the child node at the split point.
   *
   * prevLeft and prevRight are used in the recursion
   *
   * Returns an object:
   *   left:  A node containing the left side of the tree.
   *   right: A node containing the right side of the tree.
   */
  splitTree: function(root, leaf, leafOffset, prevLeft, prevRight)
  {
    var left, right;

    switch (leaf.nodeType)
    {
      case 1:
      case 11:

        if (root == leaf)
        {
          left  = this.ownerDocument.createDocumentFragment();
          right = this.ownerDocument.createDocumentFragment();
        }
        else
        {
          left  = leaf.cloneNode(false);
          right = leaf.cloneNode(false);
        }

        for (var i = 0; i < leafOffset; i++)
          left.appendChild(leaf.childNodes[i].cloneNode(true));

        var rightOffset;
        if (prevLeft && prevRight)
        {
          left.appendChild(prevLeft);
          right.appendChild(prevRight);
          rightOffset = leafOffset + 1;
        }
        else
          rightOffset = leafOffset;

        for (var i = rightOffset; i < leaf.childNodes.length; i++)
          right.appendChild(leaf.childNodes[i].cloneNode(true));
        break;
      case 3:
        var text = leaf.nodeValue;
        left  = this.ownerDocument.createTextNode(text.substr(0, leafOffset));
        right = this.ownerDocument.createTextNode(text.substr(leafOffset));
        break;
    }

    if (root == leaf)
      return {left: left, right: right, idx: leafOffset, type: leaf.nodeType};
    else
      return this.splitTree(root, leaf.parentNode, xopus_globs.utils.indexOf(leaf), left, right);
  },

  /**
   * This function intersects two sets of nodes. Each set contains a leftSplit and a rightSplit.
   * leftSplit.left and leftSplit.right merged are equal to rightSplit.left and rightSplit.right merged.
   *
   * The parts can be elements, textnodes or document fragments.
   *
   * Returns a node containing the intersection.
   */
  intersectSets: function (leftSplit, rightSplit)
  {
    var node;

    switch (leftSplit.type)
    {
      case 1:
      case 11:
        node = leftSplit.right.cloneNode(false);
        for (var i = 0; i < rightSplit.idx - leftSplit.idx; i++)
          node.appendChild(leftSplit.right.childNodes[i].cloneNode(true));

        if (rightSplit.idx > leftSplit.idx && rightSplit.idx != rightSplit.left.childNodes.length)
          node.appendChild(rightSplit.left.childNodes[rightSplit.idx].cloneNode(true))
        break;
      case 3:
        var rightText = leftSplit.right.nodeValue;
        var intersectionText = rightText.substr(0, rightSplit.idx - leftSplit.idx);
        node = this.ownerDocument.createTextNode(intersectionText);
        break;
    }

    return node;
  },

  /**
   * This function does exactly the opposite of intersectSets.
   */
  excludeSets: function (leftSplit, rightSplit)
  {
    var node;

    switch (leftSplit.type)
    {
      case 1:
      case 11:
        node = leftSplit.left.cloneNode(false);
        for (var i = 0; i < leftSplit.left.childNodes.length; i++)
          node.appendChild(leftSplit.left.childNodes[i].cloneNode(true));

        for (var i = 0; i < rightSplit.right.childNodes.length; i++)
          node.appendChild(rightSplit.right.childNodes[i].cloneNode(true));
        break;
      case 3:
        var leftText  = leftSplit.left.nodeValue;
        var rightText = rightSplit.right.nodeValue;
        node = this.ownerDocument.createTextNode(leftText + rightText);
        break;
    }

    return node;
  },

  /**
   * Insert a node before the current selection.
   */
  insertNode: function(node)
  {
    var parts = this.splitTree(this.startContainer, this.startContainer, this.startOffset);

    switch (parts.type)
    {
      case 1:
      case 11:
        xopus_globs.utils.removeChildren(this.startContainer);
        this.startContainer.appendChild(parts.left);
        this.startContainer.appendChild(node);
        this.startContainer.appendChild(parts.right);
        break;
      case 3:
        var container = this.startContainer.parentNode;

        // Insert the 3 new nodes before the text node
        container.insertBefore(parts.right, this.startContainer);
        container.insertBefore(node, parts.right);
        container.insertBefore(parts.left, (node.nodeType == 11 ? node.lastChild : node));

        // Remove the old text node
        container.removeChild(this.startContainer);
        break;
    }
  },
  
  /**
   * Convert this range to a string.
   */
  toString: function()
  {
    return "Start: " + this.startContainer.xml + " @ " + this.startOffset + "\nEnd: " + this.endContainer.xml + " @ " + this.endOffset;  
  }
};

numScriptsLoading--;
