////////////////////////////////////////////////////////////////////////////////////
//
// This Code Is Licensed Under The Xopus Open Source License
// More information: http://xopus.org, info@xopus.org
// Mailing list: http://groups.yahoo.com/group/xopus
//
// This product includes software developed by Q42 BV (http://www.q42.nl/).
// please visit http://www.q42.nl for more information or mail us: xopus@q42.nl
//
////////////////////////////////////////////////////////////////////////////////////

var SV_VALIDATION_ERROR       = 256;
var SV_MISSING_ELEMENT        = SV_VALIDATION_ERROR + 1;
var SV_EXPECTED_OTHER_ELEMENT = SV_VALIDATION_ERROR + 2;
var SV_EXPECTED_NO_ELEMENT    = SV_VALIDATION_ERROR + 3;
var SV_UNKNOWN_ELEMENT        = SV_VALIDATION_ERROR + 4;

/**
 * This class can be used to partially validate a node against a given schema.
 */
function SchemaValidator (cache)
{
  this.cache = cache;
}

/**
 * Checks if the given node is valid.
 */
SchemaValidator.prototype.isValid = function (node)
{
  var valid = false;
  try
  {
    this.validate(node, function() {throw "stop";});
    valid = true;
  }
  catch (e) 
  {
    if (e != "stop")
      __xopus_alert("An error occured while validating: " + e.description);
  }
  
  return valid;
};

/**
 * Validate the given node (assumed to be an element). 
 *
 * node           The node to validate
 * errorHandler   The handler function that will be called if the node is not valid.
 *                It has one argument: the unexpected node. 
 *                Validation can be stopped by throwing an object in the handler function.
 *                Validation can be retried by returning true.
 *
 * TODO: this implementation fails late, so the node at which it fails is not necessarily 
 *       the node where the error occured.
 *
 * TODO: retry does not work
 */
SchemaValidator.prototype.validate = function (node, errorHandler)
{
  var decl = this.cache.getDeclaration(node);  
  
  if (!decl)
    this.error(SV_UNKNOWN_ELEMENT, {node: node}, errorHandler);
  
  //var startTime = new Date();
  this.validateSingleParticle([node], 0, decl, errorHandler, "\n ");
  //__xopus_alert("Done in: " + (new Date() - startTime) + " ms.");
};

SchemaValidator.prototype.validateParticle = function (nodeList, idx, particleDecl, errorHandler)
{
  var maxOccurs = (particleDecl.maxOccurs * 1 + 1 || Infinity) - 1;
  
  var maxHandler = function () { throw maxHandler; }; 
  var currHandler = errorHandler;
  
  try
  {
    var currIdx = idx;
    for (var i = 0; i < maxOccurs; i++)
    {
      if (i == particleDecl.minOccurs)
        currHandler = maxHandler;
      
      var prevIdx = currIdx;
      currIdx = this.validateSingleParticle(nodeList, currIdx, particleDecl, currHandler);
      
      if (currIdx == prevIdx)
        break;
    }
  }
  catch (e)
  {
    if (e != maxHandler)
      throw e;
  }
  
  return currIdx;
};

SchemaValidator.prototype.validateSingleParticle = function (nodeList, idx, particleDecl, errorHandler)
{
  switch (particleDecl.itemType)
  {
    case SOMITEM_GROUP:
      return this.validateGroup(nodeList, idx, particleDecl, errorHandler);
    case SOMITEM_ALL:
      return this.validateAll(nodeList, idx, particleDecl, errorHandler);
    case SOMITEM_CHOICE:
      return this.validateChoice(nodeList, idx, particleDecl, errorHandler);
    case SOMITEM_SEQUENCE:
      return this.validateSequence(nodeList, idx, particleDecl, errorHandler);
    case SOMITEM_EMPTYPARTICLE:
      return this.validateEmptyParticle(nodeList, idx, particleDecl, errorHandler);
    case SOMITEM_ELEMENT:
      return this.validateElement(nodeList, idx, particleDecl, errorHandler);
    default:
      throw new Error("SingleParticle: unknown particleDecl.itemType: " + particleDecl.itemType);
  }
};

SchemaValidator.prototype.validateGroup = function (nodeList, idx, groupDecl, errorHandler)
{
  return this.validateSingleParticle(nodeList, idx, groupDecl.particles.item(0), errorHandler);
};

SchemaValidator.prototype.validateAll = function (nodeList, idx, groupDecl, errorHandler)
{
  __xopus_alert("Not implemented yet");
  
  return idx;
};

SchemaValidator.prototype.validateChoice = function (nodeList, idx, groupDecl, errorHandler)
{
  var optionalHandler = function(type, args){optionalHandler.type = type; optionalHandler.args = args; throw optionalHandler;};
  
  var foundValidParticle = false;
  var currIdx = idx;
  for (var i = 0; i < groupDecl.particles.length; i++)
  {
    var prevIdx = currIdx;
    
    try
    {
      currIdx = this.validateParticle(nodeList, currIdx, groupDecl.particles.item(i), optionalHandler);
    }
    catch (e)
    {
      if (e != optionalHandler)
        throw e;
      else
        continue;
    }
    
    foundValidParticle = true;
    
    if (prevIdx < currIdx)
      break;
  }
  
  // Check if we haven't found a match where we should have.
  if (!foundValidParticle && groupDecl.particles.length > 0)
  {
    this.error(optionalHandler.type, optionalHandler.args, errorHandler);
    // TODO currIdx kan je alleen zetten als node een element van nodeList is.
    currIdx = xopus_globs.utils.indexOf(optionalHandler.args.node) + 1;
  }
  
  return currIdx;
};

SchemaValidator.prototype.validateSequence = function (nodeList, idx, groupDecl, errorHandler)
{
  for (var i = 0; i < groupDecl.particles.length; i++)
    idx = this.validateParticle(nodeList, idx, groupDecl.particles.item(i), errorHandler);
    
  return idx;
};

SchemaValidator.prototype.validateEmptyParticle = function (nodeList, idx, groupDecl, errorHandler)
{
  return this.skipInsignificantNodes(nodeList, idx);
};  

SchemaValidator.prototype.validateElement = function (nodeList, idx, eltDecl, errorHandler)
{
  var elt = nodeList[idx];
  
  if (!elt)
    this.error(SV_MISSING_ELEMENT, {decl: eltDecl}, errorHandler);
  
  if (xopus_globs.utils.getLocalName(elt) != eltDecl.name || xopus_globs.utils.getNamespaceURI(elt) != eltDecl.namespaceURI)
    this.error(SV_EXPECTED_OTHER_ELEMENT, {node: elt, decl: eltDecl}, errorHandler);
  
  var lastIdx = this.validateType(elt.childNodes, 0, eltDecl.type, errorHandler);
  
  if (lastIdx < elt.childNodes.length)
    this.error(SV_EXPECTED_NO_ELEMENT, {node: elt.childNodes[lastIdx]}, errorHandler);
  
  return this.skipInsignificantNodes(nodeList, idx + 1); 
};

SchemaValidator.prototype.validateType = function (nodeList, idx, typeDecl, errorHandler)
{
  idx = this.skipInsignificantNodes(nodeList, idx);
  
  var nextIdx;
  if ((typeDecl.itemType & SOMITEM_DATATYPE) == SOMITEM_DATATYPE)
  {
    nextIdx = idx + 1;
  }
  else
  {
    switch (typeDecl.contentType)
    {
      case SCHEMACONTENTTYPE_MIXED:
      case SCHEMACONTENTTYPE_ELEMENTONLY:
        nextIdx = this.validateParticle(nodeList, idx, typeDecl.contentModel, errorHandler);
        break;
      default:
        nextIdx = idx + 1;
    }
  }
  
  return this.skipInsignificantNodes(nodeList, nextIdx);
};

SchemaValidator.prototype.error = function (type, arg, handler)
{
  handler(type, arg);
};

SchemaValidator.prototype.skipInsignificantNodes = function (nodeList, idx)
{
  while (nodeList[idx] && (nodeList[idx].nodeType == 3 || nodeList[idx].nodeType == 7))
    idx++;
  
  return idx;
};

numScriptsLoading--;